//-----------------------------------------------------------------------------
// name: hw2_2.ck
// desc: HW2 Milestone 2, currently unnamed polar flower real time audio visualizer.
// 
// author: Elisse Chow
// date: Fall 2023
//-----------------------------------------------------------------------------

// Set up camera
GG.camera() --> GGen dolly --> GG.scene();
GG.camera().posZ( 10 );
GG.scene().backgroundColor( @(0, 0, 0) );

// Define constants
0.0 => float THETA_MIN;  // Polar function minimum
8 * Math.PI => float THETA_MAX;  // Polar function maximum
1024 => int MAX_ITERS;  // Number of sampled points on the polar curve
1024 => int WINDOW_SIZE;  // Window size
256 => int WATERFALL_DEPTH;  // Waterfall depth

Mouse mouse;
// open mouse device 0
spork ~ mouse.start(0);
for ( int i : [0, 1, 2]) {
    spork ~ wait_on_button(i);
}

float samples[WINDOW_SIZE];
complex response[WINDOW_SIZE];
float time_positions[WINDOW_SIZE * 3];
vec3 positions[MAX_ITERS];
Waterfall waterfall --> GG.scene();

// waveform renderer
GLines waveform --> GG.scene(); waveform.mat().lineWidth(1.0);
// translate up
waveform.posY(-1.5);
// color0
waveform.mat().color( hsv_to_rgb(122.0, 0.93, 0.66) );

// Audio processing from sndpeek-minimal.ck

// Accumulate samples from mic.
adc => Flip accum => blackhole;
// Take the FFT
adc => PoleZero dcbloke => FFT fft => blackhole;
// Set DC blocker
.95 => dcbloke.blockZero;
// Set size of flip
WINDOW_SIZE => accum.size;
// Set window type and size
Windowing.hann(WINDOW_SIZE) => fft.window;
// Set FFT size (will automatically zero pad)
WINDOW_SIZE*2 => fft.size;
// Get a reference for our window for visual tapering of the waveform
Windowing.hann(WINDOW_SIZE) @=> float window[];

// map audio buffer to 3D positions
fun void map2waveform( float in[], float out[] )
{
    if( in.size()*3 != out.size() )
    {
        <<< "size mismatch in flatten()", "" >>>;
        return;
    }
    
    // mapping to xyz coordinate
    int i;
    3 => float width;
    for( auto s : in )
    {
        // map
        -width/2 + width/WINDOW_SIZE*i => out[i*3];
        s*6 * window[i] => out[i*3+1];
        0 => out[i*3+2];
        // increment
        i++;
    }
}

/*
    HSV to RGB converter.
    Thanks @tae kyu for the link to the calcs!
    https://www.rapidtables.com/convert/color/hsv-to-rgb.html

    Args:
        h: Hue [0, 360]
        s: Saturation [0, 1]
        v: Value [0, 1]
    Returns:
        vec3 for RGB color

*/
fun vec3 hsv_to_rgb(float h, float s, float v) {
    v * s => float c;
    (h / 60) % 2 - 1 => float temp;
    if (temp < 0) {
        -temp => temp;
    }

    c * (1 - temp) => float x;
    v - c => float m;

    vec3 rgb_prime;
    if (0 <= h && h < 60) {
        @(c, x, 0.0) => rgb_prime;
    } else if (60 <= h && h < 120) {
        @(x, c, 0.0) => rgb_prime;
    } else if (120 <= h && h < 180) {
        @(0.0, c, x) => rgb_prime;
    } else if (180 <= h && h < 240) {
        @(0.0, x, c) => rgb_prime;
    } else if (240 <= h && h < 300) {
        @(x, 0.0, c) => rgb_prime;
    } else {
        @(c, 0.0, x) => rgb_prime;
    }

    vec3 rgb;
    (rgb_prime.x + m) => rgb.x;
    (rgb_prime.y + m) => rgb.y;
    (rgb_prime.z + m) => rgb.z;
    return rgb;
}

class Waterfall extends GGen {
    0 => int playhead;  // Waterfall playhead
    GPoints wfl[WATERFALL_DEPTH];  // All points
    @(1.0, 1.0, 1.0) => vec3 color;

    // Iterate of point GGens
    for ( GPoints w : wfl) {
        w --> this;

        w.mat().color( @(1.0, 1.0, 1.0) );
    }

    fun void latest(vec3 positions[]) {
        wfl[playhead].geo().positions(positions);
        playhead++;
        WATERFALL_DEPTH %=> playhead;
    }

    fun void update(float dt) {
        playhead => int pos;
        WATERFALL_DEPTH/2 => float thresh;
        WATERFALL_DEPTH - thresh => float fadeChunk;
        for( int i; i < wfl.size(); i++ )
        {
            // Start with playhead-1 and go backwards
            pos--; if( pos < 0 ) WATERFALL_DEPTH-1 => pos;
            // Offset Z
            wfl[pos].posZ( -i/5 );
            // Color based on position in history
            wfl[pos].mat().color( hsv_to_rgb(i, 0.7, 1 - 0.5 * i$float / WATERFALL_DEPTH$float) );

            // Saving extra methods to offset + color history.
            // wfl[pos].posZ( -1.5 * Math.log(i/5 + 1) );
            // wfl[pos].mat().color( hsv_to_rgb(i, 1 - 0.5 * i$float / WATERFALL_DEPTH$float , 1 - 0.5 * i$float / WATERFALL_DEPTH$float) );
            // wfl[pos].mat().color( hsv_to_rgb(300.0, 0.8 - 2*i$float / WATERFALL_DEPTH$float, 1 - 0.5 * i$float / WATERFALL_DEPTH$float) );
        }
    }
}

/*
    Polar function for different flowers (3 choices).
    
    Args:
        theta: Calculate the radius for the given theta
        r_scale: Scale factor for the radius; used to show history
        c: Coefficient to drive animation
    Returns:
        Calculated radius

*/
fun float polar_flower(int graph_choice, float theta, float r_scale, float c) {
    if (graph_choice == 0) {  // Simple flower
        return r_scale * Math.cos(c * theta);
    } else if (graph_choice == 1) {  // Moderately complex flower
        return r_scale * Math.cos(c * theta) + Math.sin(Math.PI * theta / 4.0);
    } else if (graph_choice == 2) {
        return r_scale / 15.0 * Math.sin(c * theta) * theta;
    } else if (graph_choice == 3) {
        return r_scale / 3.0 * (Math.sin(3 * theta / 4.0 + ((1 + c) * Math.PI/6)) - Math.acos(Math.cos(6 * theta)));
    }
    return r_scale;
}

/*
    Returns true if point should be visible according to the samples.

    Args:
        theta: The angle of the position
        samples: The sampled values of the FFT
        threshold: The minimum value of a sample to show a point

    Returns:
        True (1) if point falls within a sample's bucket that has
        a value greater than the given threshold

*/
fun int show_point(float theta, complex response[], float threshold) {
    Math.floor(WINDOW_SIZE * theta/(2 * Math.PI))$int => int bucket;
    return 5 * Math.sqrt((response[bucket]$polar).mag * 10) > threshold;
}

/*
    Graph polar function given spectrum.

    Args:
        samples: FFT samples.
        out: Array to store positions
        r_scale: Scale factor for the radius; used to show history
        c: Coefficient driving the animation
    
*/
fun void graph(complex response[], vec3 out[], int graph_choice, float r_scale, float c) {
    for (int i; i < MAX_ITERS; i++) {
        i * (THETA_MAX - THETA_MIN) / MAX_ITERS => float theta;
        polar_flower(graph_choice, theta, r_scale, c) => float r;
        
        r * Math.sin(theta) => float x;
        r * Math.cos(theta) => float y;

        // Get wrapped angle so theta is in range [-PI, PI]
        Math.atan2(y, x) => float wrapped_theta;

        if (show_point(wrapped_theta, response, 0.1)) {
            x => out[i].x;
            y => out[i].y;
            0 => out[i].z;
        } else {
            0 => out[i].x;
            0 => out[i].y;
            -100 => out[i].z;
        }
    }
    waterfall.latest( out );
}

1 => int DIRECTION;
5.0 => float MAX_C;
1.5 => float MIN_C;
/*
    Update animation coefficient by incrementing or decrementing
    based on constants defined above.
    
    Args:
        c: Current value of c
    Returns:
        Updated c value
*/
fun float update_coefficient(float c, float amplitude) {
    // Switch direction of animation if over maximum c value
    if (c > MAX_C || c < MIN_C) {
        -DIRECTION => DIRECTION;
    }
    return c + DIRECTION * amplitude;
}


fun float max_value(float in[]) {
    -1000.0 => float max;
    for ( auto s : in ) {
        Math.max(max, s) => max;
    } 
    return max;
}

1 => int animated;
int graph_choice;
5 => int max_choices;
1 => int show_stem;
fun void wait_on_button(int which) {
    if( which < 0 || which > mouse.mouseDownEvents.size() )
    {
        <<< "cannot wait on button with index", which >>>;
        return;
    }
    
    // loop
    while( true )
    {
        // wait on a particular mouse button is hit
        mouse.mouseDownEvents[which] => now;
        // print
        // <<< "*** mouse BUTTON HIT", which, "***" >>>;
        if (which == 0) {  // Iterate through graph choices.
            (graph_choice + 1) % max_choices => graph_choice;
            <<< "SHOWING POLAR GRAPH", graph_choice >>>;
        } else if (which == 1) {
            if (animated) {  // Toggle animation state.
                0 => animated;
                <<< "PAUSED ANIMATION" >>>;
            } else {
                1 => animated;
                <<< "STARTED ANIMATION" >>>;
            }
        } else {  // Toggle time domain "stem"
            if (show_stem) {
                0 => show_stem;
                <<< "HIDE STEM" >>>;
            } else {
                1 => show_stem;
                <<< "SHOW STEM" >>>;
            }
        }
    }
}

fun void do_audio() {
    while (true) {
        // Upchuck to process accum
        accum.upchuck();
        accum.output( samples );
        fft.upchuck();
        fft.spectrum ( response );
        WINDOW_SIZE::samp/2 => now;
    }
}
spork ~ do_audio();

<<< "LEFT MOUSE CLICK TO CHANGE FLOWER" >>>;
<<< "RIGHT MOUSE CLICK TO STOP ANIMATION" >>>;
<<< "MIDDLE MOUSE CLICK TO TOGGLE STEM" >>>;

1.5 => float scale;
MIN_C => float c;
waveform.rotZ(Math.PI/2);
while (true) {
    // map to interleaved format
    map2waveform( samples, time_positions );
    // set the mesh position
    waveform.geo().positions( time_positions );
    if (show_stem) {
        waveform.posZ(0);
    } else {
        waveform.posZ(100);
    }
    max_value( samples ) / 50.0  => float amplitude;
    graph(response, positions, graph_choice,  scale, c);
    if (animated) {
        update_coefficient(c, amplitude) => c;
    }
    GG.nextFrame() => now;
}

